package Text::WikiFormat;

use strict;

use vars qw( $VERSION %html_tags );
$VERSION = 0.12;

%html_tags = (
    newline    => '<br />',
    line       => '<hr />',
    link       => \&make_html_link,
    strong     => sub { "<strong>$_[0]</strong>" },
    emphasized => sub { "<em>$_[0]</em>" },

    code      => [ '<code>', "</code><br>\n", '',     "\n" ],
    paragraph => [ '<p>',    "</p>\n",    '',     "\n" ],
    unordered => [ '<ul>',   "</ul>\n",   '<li>', "</li>\n" ],
    ordered   =>
    [ '<ol>', "</ol>\n", sub { qq|<li value="$_[1]">$_[0]</li>\n| } ],

    lists => {
        ordered   => qr/^([\dA-Za-z]+)\.\s*/,
        unordered => qr/^\*\s*/,
        code      => qr/^\s*/,
    },
);

sub format {
    my ( $text, $newtags, $opts ) = @_;
    $opts ||= { prefix => '', extended => 1 };

    my %tags = %html_tags;
    if ( defined $newtags and %$newtags ) {
        @tags{ keys %$newtags } = values %$newtags;
    }

    my %lists = map { $_ => [] } qw( code ordered unordered paragraph );
    my ( $parsed, $active_list ) = ( '', '' );

    for my $line ( split ( /\n/, $text ) ) {

        # list element
        if ( $line =~ s/^(?:\t+|\s{4,})// ) {

            foreach my $list ( sort { $b cmp $a } keys %{ $tags{lists} } ) {

                my $regex = $tags{lists}->{$list};
                if ( my @captures = ( $line =~ $regex ) ) {
                    $line =~ s/$regex//;

                    $line = format_line( $line, \%tags, $opts );

                    my $action = $html_tags{$list} or last;
                    my $formatted;

                    if ( @$action == 3 ) {
                        my $subref = $action->[2];
                        if ( defined $subref and defined &$subref ) {
                            $formatted = $subref->( $line, @captures );
                        }
                        else {
                            warn "Bad actions for list type '$list'\n";
                        }
                    }
                    else {
                        $formatted = $action->[2] . $line . $action->[3];
                    }

                    # save this list element
                    push @{ $lists{$list} }, $formatted;

                    # must end previous list type
                    if ( $active_list and $active_list ne $list ) {
                        $parsed .= end_list( \%lists, $active_list,
                            $html_tags{$active_list} );
                    }
                    $active_list = $list;

		    last;
                }
            }
        }
        else {
            if ( $active_list and $active_list ne 'paragraph' or !$line ) {
                pop @{ $lists{paragraph} } unless $line;
                $parsed .=
                  end_list( \%lists, $active_list, $html_tags{$active_list} );
            }

            next unless $line;
            $active_list = 'paragraph';
            push @{ $lists{paragraph} }, format_line( $line, \%tags, $opts ),
              $html_tags{newline};
        }
    }
    pop @{ $lists{paragraph} } if $active_list eq 'paragraph';
    $parsed .= end_list( \%lists, $active_list, $html_tags{$active_list} )
      if $active_list;
    return $parsed;
}

sub end_list {
    my ( $lists, $active, $tags ) = @_;

    my $result = join ( '', $tags->[0], @{ $lists->{$active} }, $tags->[1] );
    $lists->{$active} = [];
    return $result;
}

use URI::Escape;

sub format_line {
    my ( $text, $tags, $opts ) = @_;
    $opts ||= {};

    $text =~ s!'''(.+?)'''!$tags->{strong}->($1)!eg;
    $text =~ s!''(.+?)''!$tags->{emphasized}->($1)!eg;
    $text =~ s!^-{4,}!$tags->{line}!gm;

    $text =~ s/\[\[([^\]]+)\]([^\]]+)\]/pretty($1, $2)/eg if $opts->{extended};
    $text =~ s!\[([^\]]+)\]!$tags->{link}->($1, $opts, 1)!eg
      if $opts->{extended};

    $text =~ s|(?<!["/>=])\b([A-Za-z]+(?:[A-Z]\w+)+)|$tags->{link}->($1, 
		$opts)|eg;

    return $text;
}

sub pretty {
    my ( $link, $title ) = @_;

    if ( $link =~ /^\w+:[\/\w]/ ) {
        return qq|<a href="$link">$title</a>|;
    }
    else {
        $link = uri_escape( $link, "\x80-\xff" );
        return qq|<a href="wiki.pl?page=$link">$title</a>|;
    }
}

sub make_html_link {
    my ( $link, $opts, $escape ) = @_;
    $opts ||= {};

    my $title;
    $title ||= $link;
    $link = uri_escape( $link, "\x80-\xff" ) if $escape;

    my $prefix = $opts->{prefix} || '';
    return qq|<a href="$prefix$link">$title</a>|;
}

sub nice_utf8 {
    join ( "", map( ( $_ > 255 ? "&#$_;" : chr ), unpack( "U0U*", $_[0] ) ) );
}

'shamelessly adapted from the Jellybean project';
